import Monad( MonadPlus( .. ), guard )

-- guard True = return()
-- guard False = mzero

newtype StateTransMany s a = STM( s -> [(s, a)] )

instance Monad (StateTransMany s)
   where
    -- (>>=) :: StateTransMany s a -> (a -> StateTransMany s b) -> StateTransMany s b
    (STM p) >>= k = STM( \s0 -> case p s0 of
                                 [] -> []
                                 ss -> concat (map (\(s, a) ->
                                                       let (STM q) = k a
                                                       in q s) ss))
                
    -- return :: a -> StateTransMany s a
    return a = STM( \s -> [(s, a)] )

instance MonadPlus (StateTransMany s)
  where
    -- mzero :: StateTransMany s a
    mzero = STM( \s -> [] )
    -- mplus ::  StateTransMany s a ->  StateTransMany s a ->  StateTransMany s a
    (STM p) `mplus` (STM q)  =  STM( \s0 -> case p s0 of
    					[] -> q s0
    					s -> s ++ q s0 )

applySTM (STM p) s = p s

con :: a -> [a]
con a = [a] 

-- We can use mplus and mzero to implement backtracking algorithms.

type QState = ([Int], [Int], [Int])

getCols = STM( \(cols, swDiags, seDiags) ->
                con ((cols, swDiags, seDiags), cols) )
getSWDiags = STM( \(cols, swDiags, seDiags) ->
                  con ((cols, swDiags, seDiags), swDiags) )
getSEDiags = STM( \(cols, swDiags, seDiags) ->
                  con ((cols, swDiags, seDiags), seDiags) )

putCols c = STM( \(cols, swDiags, seDiags) ->
                  con ((c:cols, swDiags, seDiags), ()) )
putSWDiags sw = STM( \(cols, swDiags, seDiags) ->
                      con ((cols, sw:swDiags, seDiags), ()) )
putSEDiags se = STM( \(cols, swDiags, seDiags) ->
                      con ((cols, swDiags, se:seDiags), ()) )

tryPutCol c =
	do cols <- getCols
	   guard (c `notElem` cols)
	   putCols c
	   
tryPutSWDiag sw =
	do swDiags <- getSWDiags
	   guard (sw `notElem` swDiags)
	   putSWDiags sw
	   
tryPutSEDiag se =
	do seDiags <- getSEDiags
	   guard (se `notElem` seDiags)
	   putSEDiags se

place r c = 
	do tryPutCol c
	   tryPutSWDiag (c-r)
	   tryPutSEDiag (c+r)

tryEach :: MonadPlus m => [a] -> (a -> m b) -> m b   
tryEach [] f = mzero
tryEach (h:t) f = f h `mplus` tryEach t f

queens :: MonadPlus (StateTransMany QState) => Int -> Int -> StateTransMany QState [Int]
queens r colNum =
	if r == 0
	then getCols             -- Success, return list of columns
	else tryEach [0..colNum-1] (\c ->
		do place (r-1) c
		   queens (r-1) colNum )

q3 = map snd (applySTM (queens 3 3) ([], [], []))
q4 = map snd (applySTM (queens 4 4) ([], [], []))
q5 = map snd (applySTM (queens 5 5) ([], [], []))
q6 = map snd (applySTM (queens 6 6) ([], [], []))
q7 = map snd (applySTM (queens 7 7) ([], [], []))
q8 = map snd (applySTM (queens 8 8) ([], [], []))
